/*
 * Panel-Card Starterkit knob input driver
 *
 * Copyright (C) 2007 taskit GmbH
 */

#include <linux/module.h>
#include <linux/interrupt.h>
#include <linux/input.h>
#include <asm/arch/gpio.h>
#include <asm/mach/irq.h>
#include <asm/delay.h>
#include <asm/arch/hardware.h>

#define KNOB_PIN_PUSH  AT91_PIN_PB31
#define KNOB_PIN_LEFT  AT91_PIN_PB30
#define KNOB_PIN_RIGHT AT91_PIN_PB29

static struct input_dev *knob_input_dev;
static int last=0;
static int rightstatus=0;
static int leftstatus=0;

static irqreturn_t knob_push_interrupt(int irq, void *dev_id)
{
	if(at91_get_gpio_value(KNOB_PIN_PUSH)==0) {
		input_report_key(knob_input_dev, KEY_ENTER, 1);
		input_sync(knob_input_dev);
	} else {
		input_report_key(knob_input_dev, KEY_ENTER, 0);
		input_sync(knob_input_dev);
	}
	return IRQ_HANDLED;
}

static irqreturn_t knob_right_interrupt(int irq, void *dev_id)
{
	if(rightstatus==at91_get_gpio_value(KNOB_PIN_RIGHT))
		return IRQ_HANDLED;
	rightstatus=at91_get_gpio_value(KNOB_PIN_RIGHT);
	if(last!=KNOB_PIN_LEFT||leftstatus!=rightstatus) {
		last=KNOB_PIN_RIGHT;
		return IRQ_HANDLED;
	}
	last=0;
	input_report_key(knob_input_dev, KEY_RIGHT, 1);
	input_sync(knob_input_dev);
	input_report_key(knob_input_dev, KEY_RIGHT, 0);
	input_sync(knob_input_dev);
	return IRQ_HANDLED;
}

static irqreturn_t knob_left_interrupt(int irq, void *dev_id)
{
	if(leftstatus==at91_get_gpio_value(KNOB_PIN_LEFT))
		return IRQ_HANDLED;
	leftstatus=at91_get_gpio_value(KNOB_PIN_LEFT);
	if(last!=KNOB_PIN_RIGHT||leftstatus!=rightstatus) {
		last=KNOB_PIN_LEFT;
		return IRQ_HANDLED;
	}
	last=0;
	input_report_key(knob_input_dev, KEY_LEFT, 1);
	input_sync(knob_input_dev);
	input_report_key(knob_input_dev, KEY_LEFT, 0);
	input_sync(knob_input_dev);
	return IRQ_HANDLED;
}

static int __init knob_init(void)
{
	int err=0;
	if(!(knob_input_dev=input_allocate_device())) {
		err=-ENOMEM;
		goto out;
	};
	knob_input_dev->name="Knob input device";
	knob_input_dev->evbit[0]=BIT(EV_KEY);
	set_bit(KEY_LEFT , knob_input_dev->keybit);
	set_bit(KEY_RIGHT, knob_input_dev->keybit);
	set_bit(KEY_ENTER, knob_input_dev->keybit);
	input_register_device(knob_input_dev);
	at91_set_gpio_input(KNOB_PIN_PUSH, 0);
	at91_set_deglitch(KNOB_PIN_PUSH, 1);
	at91_set_gpio_input(KNOB_PIN_LEFT, 1);
	at91_set_deglitch(KNOB_PIN_LEFT, 1);
	at91_set_gpio_input(KNOB_PIN_RIGHT, 1);
	at91_set_deglitch(KNOB_PIN_RIGHT, 1);

	if((err=request_irq(KNOB_PIN_PUSH, knob_push_interrupt, 0, "knob-push", NULL)))
		goto err_irq;
	if((err=request_irq(KNOB_PIN_LEFT, knob_left_interrupt, 0, "knob-left", NULL)))
		goto err_irq2;
	if((err=request_irq(KNOB_PIN_RIGHT, knob_right_interrupt, 0, "knob-right", NULL)))
		goto err_irq3;
	return 0;

err_irq3:
	free_irq(KNOB_PIN_LEFT, 0);
err_irq2:
	free_irq(KNOB_PIN_PUSH, 0);
err_irq:
	input_unregister_device(knob_input_dev);
out:
	return err;
}

static void __exit knob_exit(void)
{
	input_unregister_device(knob_input_dev);
	free_irq(KNOB_PIN_PUSH, 0);
	free_irq(KNOB_PIN_LEFT, 0);
	free_irq(KNOB_PIN_RIGHT, 0);
}

module_init(knob_init)
module_exit(knob_exit)

MODULE_AUTHOR("Christian Glindkamp");
MODULE_DESCRIPTION("Input driver for the Panel-Card Starterkit knob");
MODULE_LICENSE("GPL v2");

