/*
 * Portux Panel-PC matrix keyboard submodule
 *
 * Copyright (C) 2005-2007 taskit GmbH
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/input.h>
#include <asm/uaccess.h>
#include <linux/string.h>
#include <linux/portuxinput.h>

matrixmap *matrix;

static struct input_dev *portuxmatrix;

static struct portuxinput_layerstate layer={
	.max=PORTUXMATRIX_MAX_LAYER,
};

static long iochandler(unsigned int cmd, unsigned long arg)
{
	struct matrixentry entry;
	matrixmap map;

	switch(cmd) {
	case PORTUXMATRIX_IOC_SETKEY:
	case PORTUXMATRIX_IOC_GETKEY:
	case PORTUXMATRIX_IOC_SETKEYMAP:
	case PORTUXMATRIX_IOC_SETLAYER:
	case PORTUXMATRIX_IOC_SETMAXLAYER:
		if(!matrix)
			return -ENXIO;
	}

	if((cmd==PORTUXMATRIX_IOC_SETKEY)||(cmd==PORTUXMATRIX_IOC_GETKEY)) {
		if(copy_from_user(&entry, (void __user *)arg, sizeof(struct matrixentry)))
			return -EFAULT;
		if(entry.x>PORTUXMATRIX_MAX_X||entry.y>PORTUXMATRIX_MAX_Y)
			return -EINVAL;
		if(cmd==PORTUXMATRIX_IOC_SETKEY||cmd==PORTUXMATRIX_IOC_GETKEY) {
			if(entry.layer>=PORTUXMATRIX_MAX_LAYER)
				return -EINVAL;
		}
	}

	switch(cmd) {
	case PORTUXMATRIX_IOC_SETKEY:
		memcpy(&((*matrix)[entry.layer][entry.y][entry.x]), &entry.value, sizeof(struct portuxinputevent));
		break;
	case PORTUXMATRIX_IOC_GETKEY:
		memcpy(&entry.value, &((*matrix)[entry.layer][entry.y][entry.x]), sizeof(struct portuxinputevent));
		if(copy_to_user((void __user *)arg, &entry, sizeof(struct matrixentry)))
			return -EFAULT;
		break;
	case PORTUXMATRIX_IOC_SETKEYMAP:
		if(copy_from_user(&map, (void __user *)arg, sizeof(matrixmap)))
			return -EFAULT;
		memcpy(matrix, &map, sizeof(matrixmap));
		break;
	case PORTUXMATRIX_IOC_SETLAYER:
		if((unsigned char)arg<layer.max)
			layer.cur=(unsigned char)arg;
		else
			return -EINVAL;
		break;
	case PORTUXMATRIX_IOC_SETMAXLAYER:
		if((unsigned char)arg<=PORTUXMATRIX_MAX_LAYER)
			layer.max=(unsigned char)arg;
		else
			return -EINVAL;
		break;
	default:
		return -ENOTTY;
	}
	return 0;
}

static void matrixhandler(struct matrix_t *event)
{
	if(likely(matrix)) {
#ifdef CONFIG_MACH_PANELCARD
		event->repeat = 0;
#endif
		portuxinput_report_event(portuxmatrix, &(*matrix)[layer.cur][event->y][event->x], event->repeat, &layer);
		input_sync(portuxmatrix);
	}
}

static int __init portuxmatrix_init(void)
{
	int i;
	
	if(portuxinput_get_platform_data())
		matrix=portuxinput_get_platform_data()->matrixkeymap;
	else
		matrix=0;

	if(matrix) {
		portuxmatrix=portuxinput_allocate_device();
		portuxmatrix->name="Portux matrix keyboard";
		portuxmatrix->evbit[0]=BIT(EV_KEY)|BIT(EV_REL);
		portuxmatrix->relbit[0]=BIT(REL_X)|BIT(REL_Y);
		for(i=0;i<KEY_MAX;i++)
			set_bit(i, portuxmatrix->keybit);
		input_register_device(portuxmatrix);
		portuxinput_register_matrix(&matrixhandler, &iochandler);
	}
	return 0;
}

static void __exit portuxmatrix_exit(void)
{
	portuxinput_deregister_matrix(&matrixhandler);
	input_unregister_device(portuxmatrix);
	kfree(portuxmatrix);
}

module_init(portuxmatrix_init)
module_exit(portuxmatrix_exit)

MODULE_AUTHOR("Christian Glindkamp");
MODULE_DESCRIPTION("matrix keyboard submodule for portuxinput");
MODULE_LICENSE("GPL v2");

